/**
 * This material is based upon work supported by 
 * the National Science Foundation under Grant No. 1140753.
 */

package sudoku.satreduce;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;

import org.sat4j.minisat.SolverFactory;
import org.sat4j.reader.DimacsReader;
import org.sat4j.reader.Reader;
import org.sat4j.specs.IProblem;
import org.sat4j.specs.ISolver;

import sudoku.util.Solver;
import sudoku.util.board.SudokuBoard;

/**
 * This class provides functionality for reducing sudoku problems
 * to SAT problems, solving sudoku problems as SAT problems, and
 * mapping SAT instance solutions to analogous sudoku instance
 * solutions.
 * 
 * For a better explanation of how this works, please see Teaching Problem Reduction:
 * NP-Completeness via Sudoku, by Dr. Andrea F. Lobo.
 *  
 * @author Dr. Andrea F. Lobo
 * @author Dr. Ganesh R. Baliga
 * @author Kenneth Janes
 */
public class SudokuToSATReducerSolver implements Solver
{
	static final String tempFileName = "temp.cnf";

	private final Writer writer;
	private int[][] boxIndex;
	private long variables, clauses;
	// private long clauseCount = 0;

	/* side effects of invoking function decode */
	int decodedValue;
	int decodedRow;
	int decodedCol;

	/**
	 * Solves a given sudoku problem by reducing it to a SAT problem,
	 * solving that SAT problem, and then mapping said solution back
	 * to the solution for that sudoku problem.
	 */
	public void solve (SudokuBoard board) {
		reduce(board);
		unmap (board);
	}

	private void unmap (SudokuBoard board ) {
		ISolver solver = SolverFactory.newDefault(); 
		try {
			Reader reader = new DimacsReader(solver);
			IProblem problem = reader.parseInstance(tempFileName); 
			int [] vars = problem.findModel();
			if (vars == null) 
				System.out.println ("Soduku does not have a solution");
			else {
				
				// TODO
				// Complete code here which takes the SAT solution and produces
				// the Sudoku solution.
			}
		} 
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Constructs a SudokuToSATReducerSolver which will perform its operations on
	 * the provided board.
	 * 
	 * @param board - the board to be solved or reducde
	 * @throws Exception
	 */
	public SudokuToSATReducerSolver() throws Exception
	{
		this(new PrintWriter(tempFileName));
	}

	private SudokuToSATReducerSolver(Writer writer) throws Exception
	{
		this.writer = writer;
	}

	/**
	 * Reduces the sudoku problem provided in the constructor to a SAT instance.
	 */
	private void reduce(SudokuBoard board)
	{
		int base = board.size() + 1;
		this.variables = base * base * base;
		
		this.clauses = calculateClauses(board);

		// boxIndex maps (box #, cell # in a box) to (board cell #)
		// Use with enumerate in atLeastOneInBox and in atMostOneInBox.
		
		boxIndex = new int[board.size()][board.size()];
		int[] tracker = new int[board.size()];
		for(int i = 0; i < Math.pow(board.size(), 2); i++)
		{
			boxIndex[board.getBox(i)][tracker[board.getBox(i)]] = i;
			tracker[board.getBox(i)]++;
		}
		
		write("p cnf " + variables + " " + clauses + "\n");
		
		for(int i = 0; i < board.size(); i++)
		{
			for(int value = 1; value <= board.size(); value++)
			{
				atLeastOneInRow(board, i, value);
				atMostOneInRow(board, i, value);
				atLeastOneInCol(board, i, value);
				atMostOneInCol(board, i, value);
				atLeastOneInBox(board, i, value);
				atMostOneInBox(board, i, value);
			}
		}
		
		for(int i = 0; i < Math.pow(board.size(), 2); i++)
		{
			cellHasAtLeastOneValue(board, i);
			cellHasAtMostOneValue(board, i);
		}
		cellSetClauses(board);
	}
	
	// generate a clause which says a row has at least one of some value
	private void atLeastOneInRow(SudokuBoard board, int row, int value)
	{
		
		// TODO
		
	}

	// generate a clause which says a row has at most one of some value
	private void atMostOneInRow(SudokuBoard board, int row, int value)
	{

		// TODO
		
	}

	// generate the clauses for cells that are already set to a value
	private void atLeastOneInCol(SudokuBoard board, int col, int value)
	{

		// TODO
		
	}

	// generate a clause which says a column has at most one of some value
	private void atMostOneInCol(SudokuBoard board, int col, int value)
	{

		// TODO
		
	}

	// generate a clause which says a given box has at least one of some value
	private void atLeastOneInBox(SudokuBoard board, int box, int value)
	{

		// TODO
	}

	// generate a clause which says a given box has at most one of some value
	private void atMostOneInBox(SudokuBoard board, int box, int value)
	{

		// TODO
		
	}

	// generate a clause which says a cell has at least one value
	private void cellHasAtLeastOneValue(SudokuBoard board, int cell)
	{

		// TODO
		
	}

	// generate a clause which says that a cell has at most one value
	private void cellHasAtMostOneValue(SudokuBoard board, int cell)
	{

		// TODO
		
	}

	// generate clauses for cells that are already set
	private void cellSetClauses(SudokuBoard board)
	{

		// TODO
		
	}

	// generate a clause which says a cell is set to some value
	private void cellIsSetTo(SudokuBoard board, int cell, int value)
	{
		write(enumerate(board, cell, value) + " ");
		endClause();
	}

	// enumerate a cell and value for the SAT solver format
	private int enumerate(SudokuBoard board, int cell, int value)
	{
		return encode(board, board.getRow(cell), board.getCol(cell), value);
	}

	// see enumerate function for use
	private int encode(SudokuBoard board, int row, int col, int value)
	{
		int base = board.size() + 1;
		return base * base * row + base * col + value;
	}

	// unmap some encoded value from a SAT solver to part of a solution
	private void decode (SudokuBoard board, int codedValue ) {
		int base = board.size() + 1;
		decodedValue = codedValue % base;
		codedValue /= base;
		decodedCol = codedValue % base;
		decodedRow = codedValue / base;
		if (decodedValue == 0)
			decodedValue = base - 1;

	}

	private void write(String s)
	{
		try
		{
			writer.write(s);
			writer.flush();
		}
		catch(IOException ex)
		{
			ex.printStackTrace();
		}
	}

	// SAT problem format specifies clauses end in 0
	private void endClause()
	{
		write("0\n");
	}

	// calculate the number of clauses which will be in the solution
	private long calculateClauses(SudokuBoard board)
	{
		
		// TODO
		
		return 0;   // dummy return value; replace
	}
}
